"""Stream abstract class."""

from __future__ import annotations

import abc
import copy
import datetime
import json
import typing as t
from os import PathLike
from pathlib import Path
from types import MappingProxyType

import singer_sdk._singerlib as singer
from singer_sdk import metrics
from singer_sdk.batch import Batcher
from singer_sdk.exceptions import (
    AbortedSyncFailedException,
    AbortedSyncPausedException,
    InvalidReplicationKeyException,
    InvalidStreamSortException,
    MaxRecordsLimitException,
)
from singer_sdk.helpers._batch import (
    BaseBatchFileEncoding,
    BatchConfig,
    SDKBatchMessage,
)
from singer_sdk.helpers._catalog import pop_deselected_record_properties
from singer_sdk.helpers._compat import datetime_fromisoformat
from singer_sdk.helpers._flattening import get_flattening_options
from singer_sdk.helpers._state import (
    finalize_state_progress_markers,
    get_starting_replication_value,
    get_state_partitions_list,
    get_writeable_state_dict,
    increment_state,
    is_state_non_resumable,
    log_sort_error,
    reset_state_progress_markers,
    write_replication_key_signpost,
    write_starting_replication_value,
)
from singer_sdk.helpers._typing import (
    TypeConformanceLevel,
    conform_record_data_types,
    is_datetime_type,
)
from singer_sdk.helpers._util import utc_now
from singer_sdk.mapper import RemoveRecordTransform, SameRecordTransform, StreamMap

if t.TYPE_CHECKING:
    import logging

    from singer_sdk.helpers import types
    from singer_sdk.helpers._compat import Traversable
    from singer_sdk.tap_base import Tap

# Replication methods
REPLICATION_FULL_TABLE = "FULL_TABLE"
REPLICATION_INCREMENTAL = "INCREMENTAL"
REPLICATION_LOG_BASED = "LOG_BASED"


class Stream(metaclass=abc.ABCMeta):  # noqa: PLR0904
    """Abstract base class for tap streams.

    :ivar context: Stream partition or context dictionary.

    .. versionadded:: 0.39.0
       The ``context`` attribute.
    """

    STATE_MSG_FREQUENCY = 10000
    """Number of records between state messages."""

    ABORT_AT_RECORD_COUNT: int | None = None
    """
    If set, raise `MaxRecordsLimitException` if the limit is exceeded.
    """

    TYPE_CONFORMANCE_LEVEL = TypeConformanceLevel.RECURSIVE
    """Type conformance level for this stream.

    Field types in the schema are used to convert record field values to the correct
    type.

    Available options are:

    - ``TypeConformanceLevel.NONE``: No conformance is performed.
    - ``TypeConformanceLevel.RECURSIVE``: Conformance is performed recursively through
      all nested levels in the record.
    - ``TypeConformanceLevel.ROOT_ONLY``: Conformance is performed only on the
      root level.
    """

    # Used for nested stream relationships
    parent_stream_type: type[Stream] | None = None
    """Parent stream type for this stream. If this stream is a child stream, this should
    be set to the parent stream class.
    """

    ignore_parent_replication_key: bool = False

    selected_by_default: bool = True
    """Whether this stream is selected by default in the catalog."""

    def __init__(
        self,
        tap: Tap,
        schema: str | PathLike | dict[str, t.Any] | singer.Schema | None = None,
        name: str | None = None,
    ) -> None:
        """Init tap stream.

        Args:
            tap: Singer Tap this stream belongs to.
            schema: JSON schema for records in this stream.
            name: Name of this stream.

        Raises:
            ValueError: TODO
            FileNotFoundError: TODO
        """
        if name:
            self.name: str = name
        if not self.name:
            msg = "Missing argument or class variable 'name'."
            raise ValueError(msg)

        self.logger: logging.Logger = tap.logger.getChild(self.name)
        self.metrics_logger = tap.metrics_logger
        self.tap_name: str = tap.name
        self.context: types.Context | None = None

        self._config: dict = dict(tap.config)
        self._tap = tap
        self._tap_state = tap.state
        self._tap_input_catalog: singer.Catalog | None = None
        self._stream_maps: list[StreamMap] | None = None
        self.forced_replication_method: str | None = None
        self._replication_key: str | None = None
        self._primary_keys: t.Sequence[str] | None = None
        self._state_partitioning_keys: list[str] | None = None
        self._schema_filepath: Path | Traversable | None = None
        self._metadata: singer.MetadataMapping | None = None
        self._mask: singer.SelectionMask | None = None
        self._schema: dict
        self._is_state_flushed: bool = True
        self._last_emitted_state: dict | None = None
        self._sync_costs: dict[str, int] = {}
        self.child_streams: list[Stream] = []
        if schema:
            if isinstance(schema, (PathLike, str)):
                if not Path(schema).is_file():
                    msg = f"Could not find schema file '{self.schema_filepath}'."
                    raise FileNotFoundError(msg)

                self._schema_filepath = Path(schema)
            elif isinstance(schema, dict):
                self._schema = schema
            elif isinstance(schema, singer.Schema):
                self._schema = schema.to_dict()
            else:
                msg = f"Unexpected type {type(schema).__name__} for arg 'schema'."
                raise ValueError(msg)

        if self.schema_filepath:
            self._schema = json.loads(self.schema_filepath.read_text())

        if not self.schema:
            msg = (
                f"Could not initialize schema for stream '{self.name}'. A valid schema "
                "object or filepath was not provided."
            )
            raise ValueError(msg)

    @property
    def stream_maps(self) -> list[StreamMap]:
        """Get stream transformation maps.

        The 0th item is the primary stream map. List should not be empty.

        Returns:
            A list of one or more map transformations for this stream.
        """
        if self._stream_maps:
            return self._stream_maps

        if self._tap.mapper:
            self._stream_maps = self._tap.mapper.stream_maps[self.name]
            self.logger.info(
                "Tap has custom mapper. Using %d provided map(s).",
                len(self.stream_maps),
            )
        else:
            self.logger.info(
                "No custom mapper provided for '%s'. Using SameRecordTransform.",
                self.name,
            )
            self._stream_maps = [
                SameRecordTransform(
                    stream_alias=self.name,
                    raw_schema=self.schema,
                    key_properties=self.primary_keys,
                    flattening_options=get_flattening_options(self.config),
                ),
            ]
        return self._stream_maps

    @property
    def is_timestamp_replication_key(self) -> bool:
        """Check is replication key is a timestamp.

        Developers can override to `True` in order to force this value, although this
        should not be required in most use cases since the type can generally be
        accurately detected from the JSON Schema.

        Returns:
            True if the stream uses a timestamp-based replication key.

        Raises:
            InvalidReplicationKeyException: If the schema does not contain the
                replication key.
        """
        if not self.replication_key:
            return False
        type_dict = self.schema.get("properties", {}).get(self.replication_key)
        if type_dict is None:
            msg = f"Field '{self.replication_key}' is not in schema for stream '{self.name}'"  # noqa: E501
            raise InvalidReplicationKeyException(msg)
        return is_datetime_type(type_dict)

    def get_starting_replication_key_value(
        self,
        context: types.Context | None,
    ) -> t.Any | None:  # noqa: ANN401
        """Get starting replication key.

        Will return the value of the stream's replication key when `--state` is passed.
        If no prior state exists, will return `None`.

        Developers should use this method to seed incremental processing for
        non-datetime replication keys. For datetime and date replication keys, use
        :meth:`~singer_sdk.Stream.get_starting_timestamp()`

        Args:
            context: Stream partition or context dictionary.

        Returns:
            Starting replication value.

        .. note::

           This method requires :attr:`~singer_sdk.Stream.replication_key` to be set
           to a non-null value, indicating the stream should be synced incrementally.
        """
        state = self.get_context_state(context)

        return (
            get_starting_replication_value(state)
            if self.replication_method != REPLICATION_FULL_TABLE
            else None
        )

    def get_starting_timestamp(
        self,
        context: types.Context | None,
    ) -> datetime.datetime | None:
        """Get starting replication timestamp.

        Will return the value of the stream's replication key when `--state` is passed.
        If no state exists, will return `start_date` if set, or `None` if neither
        the stream state nor `start_date` is set.

        Developers should use this method to seed incremental processing for date
        and datetime replication keys. For non-datetime replication keys, use
        :meth:`~singer_sdk.Stream.get_starting_replication_key_value()`

        Args:
            context: Stream partition or context dictionary.

        Returns:
            `start_date` from config, or state value if using timestamp replication.

        Raises:
            ValueError: If the replication value is not a valid timestamp.

        .. note::

           This method requires :attr:`~singer_sdk.Stream.replication_key` to be set
           to a non-null value, indicating the stream should be synced incrementally.
        """
        value = self.get_starting_replication_key_value(context)

        if value is None:
            return None

        if not self.is_timestamp_replication_key:
            msg = f"The replication key {self.replication_key} is not of timestamp type"
            raise ValueError(msg)

        result = datetime_fromisoformat(value)
        return result if result.tzinfo else result.replace(tzinfo=datetime.timezone.utc)

    @property
    def selected(self) -> bool:
        """Check if stream is selected.

        Returns:
            True if the stream is selected.
        """
        return self.mask.get((), True)

    @selected.setter
    def selected(self, value: bool | None) -> None:
        """Set stream selection.

        Args:
            value: True if the stream is selected.
        """
        self.metadata.root.selected = value
        self._mask = self.metadata.resolve_selection()

    @t.final
    @property
    def has_selected_descendents(self) -> bool:
        """Check descendents.

        Returns:
            True if any child streams are selected, recursively.
        """
        return any(
            child.selected or child.has_selected_descendents
            for child in self.child_streams or []
        )

    @t.final
    @property
    def descendent_streams(self) -> list[Stream]:
        """Get child streams.

        Returns:
            A list of all children, recursively.
        """
        result: list[Stream] = [*self.child_streams]
        for child in self.child_streams:
            result += child.descendent_streams or []
        return result

    def _write_replication_key_signpost(
        self,
        context: types.Context | None,
        value: datetime.datetime | str | int | float,
    ) -> None:
        """Write the signpost value, if available.

        Args:
            context: Stream partition or context dictionary.
            value: TODO

        Returns:
            TODO
        """
        if not value:
            return

        state = self.get_context_state(context)
        write_replication_key_signpost(state, value)

    def compare_start_date(self, value: str, start_date_value: str) -> str:
        """Compare a bookmark value to a start date and return the most recent value.

        If the replication key is a datetime-formatted string, this method will parse
        the value and compare it to the start date. Otherwise, the bookmark value is
        returned.

        If the tap uses a non-datetime replication key (e.g. an UNIX timestamp), the
        developer is encouraged to override this method to provide custom logic for
        comparing the bookmark value to the start date.

        Args:
            value: The replication key value.
            start_date_value: The start date value from the config.

        Returns:
            The most recent value between the bookmark and start date.
        """
        if self.is_timestamp_replication_key:
            return max(value, start_date_value, key=datetime_fromisoformat)

        return value

    def _write_starting_replication_value(self, context: types.Context | None) -> None:
        """Write the starting replication value, if available.

        Args:
            context: Stream partition or context dictionary.
        """
        value = None
        state = self.get_context_state(context)

        if self.replication_key:
            replication_key_value = state.get("replication_key_value")
            if replication_key_value and self.replication_key == state.get(
                "replication_key",
            ):
                value = replication_key_value

            # Use start_date if it is more recent than the replication_key state
            start_date_value: str | None = self.config.get("start_date")
            if start_date_value:
                if not value:
                    value = start_date_value
                else:
                    value = self.compare_start_date(value, start_date_value)

        write_starting_replication_value(state, value)

    def get_replication_key_signpost(
        self,
        context: types.Context | None,  # noqa: ARG002
    ) -> datetime.datetime | t.Any | None:  # noqa: ANN401
        """Get the replication signpost.

        For timestamp-based replication keys, this defaults to `utc_now()`. For
        non-timestamp replication keys, default to `None`. For consistency in subsequent
        calls, the value will be frozen (cached) at its initially called state, per
        partition argument if applicable.

        Developers may optionally override this method in advanced use cases such
        as unsorted incremental streams or complex hierarchical stream scenarios.
        For more info: :doc:`/implementation/state`

        Args:
            context: Stream partition or context dictionary.

        Returns:
            Max allowable bookmark value for this stream's replication key.
        """
        return utc_now() if self.is_timestamp_replication_key else None

    @property
    def schema_filepath(self) -> Path | Traversable | None:
        """Get path to schema file.

        Returns:
            Path to a schema file for the stream or `None` if n/a.
        """
        return self._schema_filepath

    @property
    def schema(self) -> dict:
        """Get schema.

        Returns:
            JSON Schema dictionary for this stream.
        """
        return self._schema

    @property
    def primary_keys(self) -> t.Sequence[str] | None:
        """Get primary keys.

        Returns:
            A list of primary key(s) for the stream.
        """
        return self._primary_keys or []

    @primary_keys.setter
    def primary_keys(self, new_value: t.Sequence[str] | None) -> None:
        """Set primary key(s) for the stream.

        Args:
            new_value: TODO
        """
        self._primary_keys = new_value

    @property
    def state_partitioning_keys(self) -> list[str] | None:
        """Get state partition keys.

        If not set, a default partitioning will be inherited from the stream's context.
        If an empty list is set (`[]`), state will be held in one bookmark per stream.

        Returns:
            Partition keys for the stream state bookmarks.
        """
        return self._state_partitioning_keys

    @state_partitioning_keys.setter
    def state_partitioning_keys(self, new_value: list[str] | None) -> None:
        """Set partition keys for the stream state bookmarks.

        If not set, a default partitioning will be inherited from the stream's context.
        If an empty list is set (`[]`), state will be held in one bookmark per stream.

        Args:
            new_value: the new list of keys
        """
        self._state_partitioning_keys = new_value

    @property
    def replication_key(self) -> str | None:
        """Get replication key.

        Returns:
            Replication key for the stream.
        """
        return self._replication_key or None

    @replication_key.setter
    def replication_key(self, new_value: str | None) -> None:
        """Set replication key for the stream.

        Args:
            new_value: TODO
        """
        self._replication_key = new_value

    @property
    def is_sorted(self) -> bool:
        """Expect stream to be sorted.

        When `True`, incremental streams will attempt to resume if unexpectedly
        interrupted.

        Returns:
            `True` if stream is sorted. Defaults to `False`.
        """
        return False

    @property
    def check_sorted(self) -> bool:
        """Check if stream is sorted.

        This setting enables additional checks which may trigger
        `InvalidStreamSortException` if records are found which are unsorted.

        Returns:
            `True` if sorting is checked. Defaults to `True`.
        """
        return True

    @property
    def metadata(self) -> singer.MetadataMapping:
        """Get stream metadata.

        Metadata attributes (`inclusion`, `selected`, etc.) are part of the Singer spec.

        Metadata from an input catalog will override standard metadata.

        Returns:
            A mapping from property breadcrumbs to metadata objects.
        """
        if self._metadata is not None:
            return self._metadata

        if self._tap_input_catalog:
            catalog_entry = self._tap_input_catalog.get_stream(self.tap_stream_id)
            if catalog_entry:
                self._metadata = catalog_entry.metadata
                return self._metadata

        self._metadata = singer.MetadataMapping.get_standard_metadata(
            schema=self.schema,
            replication_method=self.forced_replication_method,
            key_properties=self.primary_keys or [],
            valid_replication_keys=(
                [self.replication_key] if self.replication_key else None
            ),
            schema_name=None,
            selected_by_default=self.selected_by_default,
        )

        # If there's no input catalog, select all streams
        self._metadata.root.selected = (
            self._tap_input_catalog is None and self.selected_by_default
        )

        return self._metadata

    @property
    def _singer_catalog_entry(self) -> singer.CatalogEntry:
        """Return catalog entry as specified by the Singer catalog spec.

        Returns:
            TODO
        """
        return singer.CatalogEntry(
            tap_stream_id=self.tap_stream_id,
            stream=self.name,
            schema=singer.Schema.from_dict(self.schema),
            metadata=self.metadata,
            key_properties=self.primary_keys or [],
            replication_key=self.replication_key,
            replication_method=self.replication_method,
            is_view=None,
            database=None,
            table=None,
            row_count=None,
            stream_alias=None,
        )

    @property
    def _singer_catalog(self) -> singer.Catalog:
        """TODO.

        Returns:
            TODO
        """
        return singer.Catalog([(self.tap_stream_id, self._singer_catalog_entry)])

    @property
    def config(self) -> t.Mapping[str, t.Any]:
        """Get stream configuration.

        Returns:
            A frozen (read-only) config dictionary map.
        """
        return MappingProxyType(self._config)

    @property
    def tap_stream_id(self) -> str:
        """Return a unique stream ID.

        Default implementations will return `self.name` but this behavior may be
        overridden if required by the developer.

        Returns:
            Unique stream ID.
        """
        return self.name

    @property
    def replication_method(self) -> str:
        """Get replication method.

        Returns:
            Replication method to be used for this stream.
        """
        if self.forced_replication_method:
            return str(self.forced_replication_method)
        if self.replication_key:
            return REPLICATION_INCREMENTAL
        return REPLICATION_FULL_TABLE

    # State properties:

    @property
    def tap_state(self) -> dict:
        """Return a writeable state dict for the entire tap.

        Note: This dictionary is shared (and writable) across all streams.

        This method is internal to the SDK and should not need to be overridden.
        Developers may access this property but this is not recommended except in
        advanced use cases. Instead, developers should access the latest stream
        replication key values using :meth:`~singer_sdk.Stream.get_starting_timestamp()`
        for timestamp keys, or
        :meth:`~singer_sdk.Stream.get_starting_replication_key_value()` for
        non-timestamp keys.

        Returns:
            A writeable state dict for the entire tap.
        """
        return self._tap_state

    def get_context_state(self, context: types.Context | None) -> dict:
        """Return a writable state dict for the given context.

        Gives a partitioned context state if applicable; else returns stream state.
        A blank state will be created in none exists.

        This method is internal to the SDK and should not need to be overridden.
        Developers may access this property but this is not recommended except in
        advanced use cases. Instead, developers should access the latest stream
        replication key values using
        :meth:`~singer_sdk.Stream.get_starting_timestamp()` for timestamp keys, or
        :meth:`~singer_sdk.Stream.get_starting_replication_key_value()` for
        non-timestamp keys.

        Partition level may be overridden by
        :attr:`~singer_sdk.Stream.state_partitioning_keys` if set.

        Args:
            context: Stream partition or context dictionary.

        Returns:
            A partitioned context state if applicable; else returns stream state.
            A blank state will be created in none exists.
        """
        state_partition_context = self._get_state_partition_context(context)
        if state_partition_context:
            return get_writeable_state_dict(
                self.tap_state,
                self.name,
                state_partition_context=state_partition_context,
            )
        return self.stream_state

    @property
    def stream_state(self) -> dict:
        """Get writable state.

        This method is internal to the SDK and should not need to be overridden.
        Developers may access this property but this is not recommended except in
        advanced use cases. Instead, developers should access the latest stream
        replication key values using :meth:`~singer_sdk.Stream.get_starting_timestamp()`
        for timestamp keys, or
        :meth:`~singer_sdk.Stream.get_starting_replication_key_value()` for
        non-timestamp keys.

        A blank state entry will be created if one doesn't already exist.

        Returns:
            A writable state dict for this stream.
        """
        return get_writeable_state_dict(self.tap_state, self.name)

    # Partitions

    @property
    def partitions(self) -> list[types.Context] | None:
        """Get stream partitions.

        Developers may override this property to provide a default partitions list.

        By default, this method returns a list of any partitions which are already
        defined in state, otherwise None.

        Returns:
            A list of partition key dicts (if applicable), otherwise `None`.
        """
        result: list[types.Mapping] = [
            partition_state["context"]
            for partition_state in (
                get_state_partitions_list(self.tap_state, self.name) or []
            )
        ]
        return result or None

    # Private bookmarking methods

    def _increment_stream_state(
        self,
        latest_record: types.Record,
        *,
        context: types.Context | None = None,
    ) -> None:
        """Update state of stream or partition with data from the provided record.

        Raises `InvalidStreamSortException` is `self.is_sorted = True` and unsorted data
        is detected.

        Note: The default implementation does not advance any bookmarks unless
        `self.replication_method == 'INCREMENTAL'.

        Args:
            latest_record: TODO
            context: Stream partition or context dictionary.

        Raises:
            ValueError: TODO
        """
        # This also creates a state entry if one does not yet exist:
        state_dict = self.get_context_state(context)

        # Advance state bookmark values if applicable
        if latest_record and self.replication_method == REPLICATION_INCREMENTAL:
            if not self.replication_key:
                msg = (
                    f"Could not detect replication key for '{self.name}' "
                    f"stream(replication method={self.replication_method})"
                )
                raise ValueError(msg)
            treat_as_sorted = self.is_sorted
            if not treat_as_sorted and self.state_partitioning_keys is not None:
                # Streams with custom state partitioning are not resumable.
                treat_as_sorted = False
            increment_state(
                state_dict,
                replication_key=self.replication_key,
                latest_record=latest_record,
                is_sorted=treat_as_sorted,
                check_sorted=self.check_sorted,
            )

    # Private message authoring methods:

    def _write_state_message(self) -> None:
        """Write out a STATE message with the latest state."""
        if (not self._is_state_flushed) and (
            self.tap_state != self._last_emitted_state
        ):
            self._tap.write_message(singer.StateMessage(value=self.tap_state))
            self._last_emitted_state = copy.deepcopy(self.tap_state)
            self._is_state_flushed = True

    def _generate_schema_messages(
        self,
    ) -> t.Generator[singer.SchemaMessage, None, None]:
        """Generate schema messages from stream maps.

        Yields:
            Schema message objects.
        """
        bookmark_keys = [self.replication_key] if self.replication_key else None
        for stream_map in self.stream_maps:
            if isinstance(stream_map, RemoveRecordTransform):
                # Don't emit schema if the stream's records are all ignored.
                continue

            yield singer.SchemaMessage(
                stream_map.stream_alias,
                stream_map.transformed_schema,
                stream_map.transformed_key_properties,
                bookmark_keys,
            )

    def _write_schema_message(self) -> None:
        """Write out a SCHEMA message with the stream schema."""
        for schema_message in self._generate_schema_messages():
            self._tap.write_message(schema_message)

    @property
    def mask(self) -> singer.SelectionMask:
        """Get a boolean mask for stream and property selection.

        Returns:
            A mapping of breadcrumbs to boolean values, representing stream and field
            selection.
        """
        if self._mask is None:
            self._mask = self.metadata.resolve_selection()
        return self._mask

    def _generate_record_messages(
        self,
        record: types.Record,
    ) -> t.Generator[singer.RecordMessage, None, None]:
        """Write out a RECORD message.

        Args:
            record: A single stream record.

        Yields:
            Record message objects.
        """
        pop_deselected_record_properties(record, self.schema, self.mask)
        record = conform_record_data_types(
            stream_name=self.name,
            record=record,
            schema=self.schema,
            level=self.TYPE_CONFORMANCE_LEVEL,
            logger=self.logger,
        )
        for stream_map in self.stream_maps:
            mapped_record = stream_map.transform(record)
            # Emit record if not filtered
            if mapped_record is not None:
                yield singer.RecordMessage(
                    stream=stream_map.stream_alias,
                    record=mapped_record,
                    version=None,
                    time_extracted=utc_now(),
                )

    def _write_record_message(self, record: types.Record) -> None:
        """Write out a RECORD message.

        Args:
            record: A single stream record.
        """
        for record_message in self._generate_record_messages(record):
            self._tap.write_message(record_message)

        self._is_state_flushed = False

    def _write_batch_message(
        self,
        encoding: BaseBatchFileEncoding,
        manifest: list[str],
    ) -> None:
        """Write out a BATCH message.

        Args:
            encoding: The encoding to use for the batch.
            manifest: A list of filenames for the batch.
        """
        self._tap.write_message(
            SDKBatchMessage(
                stream=self.name,
                encoding=encoding,
                manifest=manifest,
            ),
        )
        self._is_state_flushed = False

    def _log_metric(self, point: metrics.Point) -> None:
        """Log a single measurement.

        Args:
            point: A single measurement value.
        """
        metrics.log(self.metrics_logger, point=point)

    def log_sync_costs(self) -> None:
        """Log a summary of Sync costs.

        The costs are calculated via `calculate_sync_cost`.
        This method can be overridden to log results in a custom
        format. It is only called once at the end of the life of
        the stream.
        """
        if len(self._sync_costs) > 0:
            msg = f"Total Sync costs for stream {self.name}: {self._sync_costs}"
            self.logger.info(msg)

    def _check_max_record_limit(self, current_record_index: int) -> None:
        """Raise an exception if dry run record limit exceeded.

        Raised if we find dry run record limit exceeded,
        aka `current_record_index > self.ABORT_AT_RECORD_COUNT - 1`.

        Args:
            current_record_index: The zero-based index of the current record.
        """
        if (
            self.ABORT_AT_RECORD_COUNT is not None
            and current_record_index > self.ABORT_AT_RECORD_COUNT - 1
        ):
            self._abort_sync(
                abort_reason=MaxRecordsLimitException(
                    "Stream prematurely aborted due to the stream's max dry run "
                    f"record limit ({self.ABORT_AT_RECORD_COUNT}) being reached.",
                ),
            )

    def _abort_sync(self, abort_reason: Exception) -> None:
        """Handle a sync operation being aborted.

        This method will attempt to close out the sync operation as gracefully as
        possible - for instance, if a max runtime or record count is reached. This can
        also be called for `SIGTERM` and KeyboardInterrupt events.

        If a state message is pending, we will attempt to write it to STDOUT before
        shutting down.

        If the stream can reach a valid resumable state, then we will raise
        `AbortedSyncPausedException`. Otherwise, `AbortedSyncFailedException` will be
        raised.

        Args:
            abort_reason: The exception that triggered the sync to be aborted.

        Raises:
            AbortedSyncFailedException: Raised if sync could not reach a valid state.
            AbortedSyncPausedException: Raised if sync was able to be transitioned into
                a valid state without data loss or corruption.
        """
        self._write_state_message()  # Write out state message if pending.

        if self.replication_method == "FULL_TABLE":
            msg = "Sync operation aborted for stream in 'FULL_TABLE' replication mode."
            raise AbortedSyncFailedException(msg) from abort_reason

        if is_state_non_resumable(self.stream_state):
            msg = "Sync operation aborted and state is not in a resumable state."
            raise AbortedSyncFailedException(msg) from abort_reason

        # Else, the sync operation can be assumed to be in a valid resumable state.
        raise AbortedSyncPausedException from abort_reason

    # Handle interim stream state

    def reset_state_progress_markers(self, state: dict | None = None) -> None:
        """Reset progress markers. If all=True, all state contexts will be set.

        This method is internal to the SDK and should not need to be overridden.

        Args:
            state: State object to promote progress markers with.
        """
        if state is None or state == {}:
            context: types.Context | None
            for context in self.partitions or [{}]:
                state = self.get_context_state(context or None)
                reset_state_progress_markers(state)
            return

        reset_state_progress_markers(state)

    def _finalize_state(self, state: dict | None = None) -> None:
        """Reset progress markers and state flushed flag to ensure state is written.

        Args:
            state: State object to promote progress markers with.
        """
        state = finalize_state_progress_markers(state)  # type: ignore[arg-type]
        self._is_state_flushed = False

    def finalize_state_progress_markers(self, state: dict | None = None) -> None:
        """Reset progress markers and emit state message if necessary.

        This method is internal to the SDK and should not need to be overridden.

        Args:
            state: State object to promote progress markers with.
        """
        if state is None or state == {}:
            for child_stream in self.child_streams or []:
                child_stream.finalize_state_progress_markers()

            context: types.Context | None
            for context in self.partitions or [{}]:
                state = self.get_context_state(context or None)
                self._finalize_state(state)
        else:
            self._finalize_state(state)

        self._write_state_message()

    # Private sync methods:

    def _process_record(
        self,
        record: types.Record,
        child_context: types.Context | None = None,
        partition_context: types.Context | None = None,
    ) -> None:
        """Process a record.

        Args:
            record: The record to process.
            child_context: The child context.
            partition_context: The partition context.
        """
        partition_context = partition_context or {}
        for key, val in partition_context.items():
            # Add state context to records if not already present
            if key not in record:
                record[key] = val

        for context in self.generate_child_contexts(
            record=record,
            context=child_context,
        ):
            # Sync children, except when primary mapper filters out the record
            if self.stream_maps[0].get_filter_result(record):
                self._sync_children(copy.copy(context))

    def _sync_records(  # noqa: C901
        self,
        context: types.Context | None = None,
        *,
        write_messages: bool = True,
    ) -> t.Generator[dict, t.Any, t.Any]:
        """Sync records, emitting RECORD and STATE messages.

        Args:
            context: Stream partition or context dictionary.
            write_messages: Whether to write Singer messages to stdout.

        Raises:
            InvalidStreamSortException: Raised if sorting errors are found while
                syncing the records.

        Yields:
            Each record from the source.
        """
        # Initialize metrics
        record_counter = metrics.record_counter(self.name)
        timer = metrics.sync_timer(self.name)

        record_index = 0
        context_element: types.Context | None
        context_list: list[types.Context] | None
        context_list = [context] if context is not None else self.partitions
        selected = self.selected

        with record_counter, timer:
            for context_element in context_list or [{}]:
                record_counter.context = context_element
                timer.context = context_element

                current_context = context_element or None
                state = self.get_context_state(current_context)
                state_partition_context = self._get_state_partition_context(
                    current_context,
                )
                self._write_starting_replication_value(current_context)
                child_context: types.Context | None = (
                    None if current_context is None else copy.copy(current_context)
                )

                for idx, record_result in enumerate(self.get_records(current_context)):
                    self._check_max_record_limit(current_record_index=record_index)

                    if isinstance(record_result, tuple):
                        # Tuple items should be the record and the child context
                        record, child_context = record_result
                    else:
                        record = record_result
                    try:
                        self._process_record(
                            record,
                            child_context=child_context,
                            partition_context=state_partition_context,
                        )
                    except InvalidStreamSortException as ex:  # pragma: no cover
                        log_sort_error(
                            log_fn=self.logger.error,
                            ex=ex,
                            record_count=record_index + 1,
                            partition_record_count=idx + 1,
                            current_context=current_context,
                            state_partition_context=state_partition_context,
                            stream_name=self.name,
                        )
                        raise

                    if selected:
                        if write_messages:
                            self._write_record_message(record)

                        self._increment_stream_state(record, context=current_context)
                        if (
                            record_index + 1
                        ) % self.STATE_MSG_FREQUENCY == 0 and write_messages:
                            self._write_state_message()

                        record_counter.increment()
                        yield record

                    record_index += 1

                if current_context == state_partition_context:
                    # Finalize per-partition state only if 1:1 with context
                    self._finalize_state(state)

        if not context:
            # Finalize total stream only if we have the full context.
            # Otherwise will be finalized by tap at end of sync.
            self._finalize_state(self.stream_state)

        if write_messages:
            # Write final state message if we haven't already
            self._write_state_message()

    def _sync_batches(
        self,
        batch_config: BatchConfig,
        context: types.Context | None = None,
    ) -> None:
        """Sync batches, emitting BATCH messages.

        Args:
            batch_config: The batch configuration.
            context: Stream partition or context dictionary.
        """
        with metrics.batch_counter(self.name, context=context) as counter:
            for encoding, manifest in self.get_batches(batch_config, context):
                counter.increment()
                self._write_batch_message(encoding=encoding, manifest=manifest)
                self._write_state_message()

    # Public methods ("final", not recommended to be overridden)

    @t.final
    def sync(self, context: types.Context | None = None) -> None:
        """Sync this stream.

        This method is internal to the SDK and should not need to be overridden.

        Args:
            context: Stream partition or context dictionary.

        Raises:
            Exception: Any exception raised by the sync process.
        """
        msg = f"Beginning {self.replication_method.lower()} sync of '{self.name}'"
        if context:
            msg += f" with context: {context}"
        self.logger.info("%s...", msg)
        self.context = MappingProxyType(context) if context else None

        # Use a replication signpost, if available
        signpost = self.get_replication_key_signpost(context)
        if signpost:
            self._write_replication_key_signpost(context, signpost)

        # Send a SCHEMA message to the downstream target:
        if self.selected:
            self._write_schema_message()

        try:
            batch_config = self.get_batch_config(self.config)
            if batch_config:
                self._sync_batches(batch_config, context=context)
            else:
                # Sync the records themselves:
                for _ in self._sync_records(context=context):
                    pass
        except Exception:
            self.logger.exception(
                "An unhandled error occurred while syncing '%s'",
                self.name,
            )
            raise

    def _sync_children(self, child_context: types.Context | None) -> None:
        if child_context is None:
            self.logger.warning(
                "Context for child streams of '%s' is null, "
                "skipping sync of any child streams",
                self.name,
            )
            return

        for child_stream in self.child_streams:
            if child_stream.selected or child_stream.has_selected_descendents:
                child_stream.sync(context=child_context)

    # Overridable Methods

    def apply_catalog(self, catalog: singer.Catalog) -> None:
        """Apply a catalog dict, updating any settings overridden within the catalog.

        Developers may override this method in order to introduce advanced catalog
        parsing, or to explicitly fail on advanced catalog customizations which
        are not supported by the tap.

        Args:
            catalog: Catalog object passed to the tap. Defines schema, primary and
                replication keys, as well as selection metadata.
        """
        self._tap_input_catalog = catalog

        catalog_entry = catalog.get_stream(self.name)
        if catalog_entry:
            self.primary_keys = catalog_entry.key_properties
            self.replication_key = catalog_entry.replication_key
            if catalog_entry.replication_method:
                self.forced_replication_method = catalog_entry.replication_method

    def _get_state_partition_context(
        self,
        context: types.Context | None,
    ) -> types.Context | None:
        """Override state handling if Stream.state_partitioning_keys is specified.

        Args:
            context: Stream partition or context dictionary.

        Returns:
            TODO
        """
        if context is None:
            return None

        if self.state_partitioning_keys is None:
            return context

        return {k: v for k, v in context.items() if k in self.state_partitioning_keys}

    def get_child_context(
        self,
        record: types.Record,
        context: types.Context | None,
    ) -> types.Context | None:
        """Return a child context object from the record and optional provided context.

        By default, will return context if provided and otherwise the record dict.

        Developers may override this behavior to send specific information to child
        streams for context.

        Return ``None`` if no child streams should be synced, for example if the
        parent record was deleted and the child records can no longer be synced.

        Args:
            record: Individual record in the stream.
            context: Stream partition or context dictionary.

        Returns:
            A dictionary with context values for a child stream, or None if no child
            streams should be synced.

        Raises:
            NotImplementedError: If the stream has children but this method is not
                overridden.
        """
        if context is None:
            for child_stream in self.child_streams:
                if child_stream.state_partitioning_keys is None:
                    parent_type = type(self).__name__
                    child_type = type(child_stream).__name__
                    msg = (
                        "No child context behavior was defined between parent stream "
                        f"'{self.name}' and child stream '{child_stream.name}'. "
                        f"The parent stream must define "
                        f"`{parent_type}.get_child_context()` and/or the child stream "
                        f"must define `{child_type}.state_partitioning_keys`."
                    )
                    raise NotImplementedError(msg)

        return context or record

    def generate_child_contexts(
        self,
        record: types.Record,
        context: types.Context | None,
    ) -> t.Iterable[types.Context | None]:
        """Generate child contexts.

        Args:
            record: Individual record in the stream.
            context: Stream partition or context dictionary.

        Yields:
            A child context for each child stream.
        """
        yield self.get_child_context(record=record, context=context)

    # Abstract Methods

    @abc.abstractmethod
    def get_records(
        self,
        context: types.Context | None,
    ) -> t.Iterable[dict | tuple[dict, dict | None]]:
        """Abstract record generator function. Must be overridden by the child class.

        Each record emitted should be a dictionary of property names to their values.
        Returns either a record dict or a tuple: (record_dict, child_context)

        A method which should retrieve data from the source and return records
        incrementally using the python `yield` operator.

        Only custom stream types need to define this method. REST and GraphQL streams
        should instead use the class-specific methods for REST or GraphQL, respectively.

        This method takes an optional `context` argument, which can be safely ignored
        unless the stream is a child stream or requires partitioning.
        More info: :doc:`/partitioning`.

        Parent streams can optionally return a tuple, in which
        case the second item in the tuple being a `child_context` dictionary for the
        stream's `context`.

        If the child context object in the tuple is ``None``, the child streams will
        be skipped. This is useful for cases where the parent record was deleted and
        the child records can no longer be synced.

        More info: :doc:`/parent_streams`

        Args:
            context: Stream partition or context dictionary.
        """

    def get_batch_config(self, config: t.Mapping) -> BatchConfig | None:  # noqa: PLR6301
        """Return the batch config for this stream.

        Args:
            config: Tap configuration dictionary.

        Returns:
            Batch config for this stream.
        """
        raw = config.get("batch_config")
        return BatchConfig.from_dict(raw) if raw else None

    def get_batches(
        self,
        batch_config: BatchConfig,
        context: types.Context | None = None,
    ) -> t.Iterable[tuple[BaseBatchFileEncoding, list[str]]]:
        """Batch generator function.

        Developers are encouraged to override this method to customize batching
        behavior for databases, bulk APIs, etc.

        Args:
            batch_config: Batch config for this stream.
            context: Stream partition or context dictionary.

        Yields:
            A tuple of (encoding, manifest) for each batch.
        """
        batcher = Batcher(
            tap_name=self.tap_name,
            stream_name=self.name,
            batch_config=batch_config,
        )
        records = self._sync_records(context, write_messages=False)
        for manifest in batcher.get_batches(records=records):
            yield batch_config.encoding, manifest

    def post_process(  # noqa: PLR6301
        self,
        row: types.Record,
        context: types.Context | None = None,  # noqa: ARG002
    ) -> dict | None:
        """As needed, append or transform raw data to match expected structure.

        Optional. This method gives developers an opportunity to "clean up" the results
        prior to returning records to the downstream tap - for instance: cleaning,
        renaming, or appending properties to the raw record result returned from the
        API.

        Developers may also return `None` from this method to filter out
        invalid or not-applicable records from the stream.

        Args:
            row: Individual record in the stream.
            context: Stream partition or context dictionary.

        Returns:
            The resulting record dict, or `None` if the record should be excluded.
        """
        return row



"""Batching utilities for Singer SDK."""

from __future__ import annotations

import itertools
import typing as t
import warnings
from abc import ABC, abstractmethod

from singer_sdk.helpers._compat import entry_points

if t.TYPE_CHECKING:
    from singer_sdk.helpers._batch import BatchConfig

_T = t.TypeVar("_T")


def __getattr__(name: str) -> t.Any:  # noqa: ANN401 # pragma: no cover
    if name == "JSONLinesBatcher":
        warnings.warn(
            "The class JSONLinesBatcher was moved to singer_sdk.contrib.batch_encoder_jsonl.",  # noqa: E501
            DeprecationWarning,
            stacklevel=2,
        )

        from singer_sdk.contrib.batch_encoder_jsonl import JSONLinesBatcher  # noqa: PLC0415, I001

        return JSONLinesBatcher

    msg = f"module {__name__} has no attribute {name}"
    raise AttributeError(msg)


def lazy_chunked_generator(
    iterable: t.Iterable[_T],
    chunk_size: int,
) -> t.Generator[t.Iterator[_T], None, None]:
    """Yield a generator for each chunk of the given iterable.

    Args:
        iterable: The iterable to chunk.
        chunk_size: The size of each chunk.

    Yields:
        A generator for each chunk of the given iterable.
    """
    iterator = iter(iterable)
    while True:
        chunk = list(itertools.islice(iterator, chunk_size))
        if not chunk:
            break
        yield iter(chunk)


class BaseBatcher(ABC):
    """Base Record Batcher."""

    def __init__(
        self,
        tap_name: str,
        stream_name: str,
        batch_config: BatchConfig,
    ) -> None:
        """Initialize the batcher.

        Args:
            tap_name: The name of the tap.
            stream_name: The name of the stream.
            batch_config: The batch configuration.
        """
        self.tap_name = tap_name
        self.stream_name = stream_name
        self.batch_config = batch_config

    @abstractmethod
    def get_batches(
        self,
        records: t.Iterator[dict],
    ) -> t.Iterator[list[str]]:
        """Yield manifest of batches.

        Args:
            records: The records to batch.

        Raises:
            NotImplementedError: If the method is not implemented.
        """
        raise NotImplementedError


class Batcher(BaseBatcher):
    """Determines batch type and then serializes batches to that format."""

    def get_batches(self, records: t.Iterator[dict]) -> t.Iterator[list[str]]:
        """Manifest of batches.

        Args:
            records: The records to batch.

        Returns:
            A list of file paths (called a manifest).
        """
        encoding_format = self.batch_config.encoding.format
        batcher_type = self.get_batcher(encoding_format)
        batcher = batcher_type(
            self.tap_name,
            self.stream_name,
            self.batch_config,
        )
        return batcher.get_batches(records)

    @classmethod
    def get_batcher(cls, name: str) -> type[BaseBatcher]:
        """Get a batcher by name.

        Args:
            name: The name of the batcher.

        Returns:
            The batcher class.

        Raises:
            ValueError: If the batcher is not found.
        """
        plugins = entry_points(group="singer_sdk.batch_encoders")

        try:
            plugin = next(filter(lambda x: x.name == name, plugins))
        except StopIteration:
            message = f"Unsupported batcher: {name}"
            raise ValueError(message) from None

        return plugin.load()  # type: ignore[no-any-return]



"""Helpers for the tap, target and mapper CLIs."""

from __future__ import annotations

import typing as t

if t.TYPE_CHECKING:
    import click

_T = t.TypeVar("_T")


class plugin_cli:  # noqa: N801
    """Decorator to create a plugin CLI."""

    def __init__(self, method: t.Callable[..., click.Command]) -> None:
        """Create a new plugin CLI.

        Args:
            method: The method to call to get the command.
        """
        self.method = method
        self.name: str | None = None

    def __get__(self, instance: _T, owner: type[_T]) -> click.Command:
        """Get the command.

        Args:
            instance: The instance of the plugin.
            owner: The plugin class.

        Returns:
            The CLI entrypoint.
        """
        return self.method(owner)



from __future__ import annotations  # noqa: A005

import datetime
import decimal
import json
import typing as t

import simplejson

__all__ = [
    "deserialize_json",
    "serialize_json",
]


def _default_encoding(obj: t.Any) -> str:  # noqa: ANN401
    """Default JSON encoder.

    Args:
        obj: The object to encode.

    Returns:
        The encoded object.
    """
    return obj.isoformat(sep="T") if isinstance(obj, datetime.datetime) else str(obj)


def deserialize_json(json_str: str | bytes, **kwargs: t.Any) -> dict:
    """Deserialize a line of json.

    Args:
        json_str: A single line of json.
        **kwargs: Optional key word arguments.

    Returns:
        A dictionary of the deserialized json.
    """
    return json.loads(  # type: ignore[no-any-return]
        json_str,
        parse_float=decimal.Decimal,
        **kwargs,
    )


def serialize_json(obj: object, **kwargs: t.Any) -> str:
    """Serialize a dictionary into a line of json.

    Args:
        obj: A Python object usually a dict.
        **kwargs: Optional key word arguments.

    Returns:
        A string of serialized json.
    """
    return simplejson.dumps(
        obj,
        use_decimal=True,
        default=_default_encoding,
        separators=(",", ":"),
        **kwargs,
    )



"""SDK for building Singer taps."""

from __future__ import annotations

from singer_sdk import streams
from singer_sdk.connectors import SQLConnector
from singer_sdk.mapper_base import InlineMapper
from singer_sdk.plugin_base import PluginBase
from singer_sdk.sinks import BatchSink, RecordSink, Sink, SQLSink
from singer_sdk.streams import GraphQLStream, RESTStream, SQLStream, Stream
from singer_sdk.tap_base import SQLTap, Tap
from singer_sdk.target_base import SQLTarget, Target

__all__ = [
    "BatchSink",
    "GraphQLStream",
    "InlineMapper",
    "PluginBase",
    "RESTStream",
    "RecordSink",
    "SQLConnector",
    "SQLSink",
    "SQLStream",
    "SQLTap",
    "SQLTarget",
    "Sink",
    "Stream",
    "Tap",
    "Target",
    "streams",
]



"""Type aliases for use in the SDK."""  # noqa: A005

from __future__ import annotations

import sys
import typing as t

import requests

if sys.version_info < (3, 9):
    from typing import Mapping  # noqa: ICN003
else:
    from collections.abc import Mapping

if sys.version_info < (3, 10):
    from typing_extensions import TypeAlias
else:
    from typing import TypeAlias  # noqa: ICN003


__all__ = [
    "Context",
    "Record",
]

Context: TypeAlias = Mapping[str, t.Any]
Record: TypeAlias = t.Dict[str, t.Any]
Auth: TypeAlias = t.Callable[[requests.PreparedRequest], requests.PreparedRequest]



"""Test Class Factory."""

from __future__ import annotations

import typing as t

import pytest

from .config import SuiteConfig
from .runners import TapTestRunner, TargetTestRunner
from .suites import (
    TestSuite,
    tap_stream_attribute_tests,
    tap_stream_tests,
    tap_tests,
    target_tests,
)

if t.TYPE_CHECKING:
    from singer_sdk import Stream, Tap, Target
    from singer_sdk.testing.templates import (
        AttributeTestTemplate,
        StreamTestTemplate,
        TapTestTemplate,
    )


class BaseTestClass:
    """Base test class."""

    params: dict[str, t.Any]
    param_ids: dict[str, list[str]]

    def __init_subclass__(cls, **kwargs: t.Any) -> None:
        """Initialize a subclass.

        Args:
            **kwargs: Keyword arguments.
        """
        # Add empty params and param_ids attributes to a direct subclass but not to
        # subclasses of subclasses
        if cls.__base__ == BaseTestClass:
            cls.params = {}
            cls.param_ids = {}


class TapTestClassFactory:
    """Factory for Tap Test Classes."""

    def __init__(
        self,
        tap_class: type[Tap],
        *,
        config: dict | None = None,
    ):
        """Initialize TapTestClassFactory.

        Args:
            tap_class: Tap class to be tested.
            config: Tap configuration for testing.
        """
        self.tap_class = tap_class
        self.config = config

    def new_test_class(
        self,
        *,
        include_tap_tests: bool = True,
        include_stream_tests: bool = True,
        include_stream_attribute_tests: bool = True,
        custom_suites: list | None = None,
        suite_config: SuiteConfig | None = None,
        **kwargs: t.Any,
    ) -> type[BaseTestClass]:
        """Get a new test class.

        Args:
            include_tap_tests: Include tap tests in the test class.
            include_stream_tests: Include stream tests in the test class.
            include_stream_attribute_tests:
                Include stream attribute tests in the test class.
            custom_suites: List of custom test suites to include in the test class.
            suite_config: SuiteConfig instance to be used when instantiating tests.
            kwargs: Default arguments to be passed to tap on create.

        Returns:
            A new test class.
        """
        # compile test suites
        suites = custom_suites or []
        if include_tap_tests:
            suites.append(tap_tests)
        if include_stream_tests:
            suites.append(tap_stream_tests)
        if include_stream_attribute_tests:
            suites.append(tap_stream_attribute_tests)

        # set default values
        if "parse_env_config" not in kwargs:
            kwargs["parse_env_config"] = True

        # create singleton test runner
        test_runner = TapTestRunner(
            tap_class=self.tap_class,
            config=self.config,
            suite_config=suite_config,
            **kwargs,
        )

        empty_test_class = self._get_empty_test_class(
            test_runner=test_runner,
            suite_config=suite_config,
        )
        return self._annotate_test_class(
            empty_test_class=empty_test_class,
            test_suites=suites,
            test_runner=test_runner,
        )

    def _get_empty_test_class(  # noqa: PLR6301
        self,
        test_runner: TapTestRunner,
        suite_config: SuiteConfig | None,
    ) -> type[BaseTestClass]:
        """Get an empty test class.

        Args:
            test_runner: Test runner to be used in the test class.
            suite_config: SuiteConfig instance to be used when instantiating tests.

        Returns:
            An empty test class.
        """

        class TapTestClass(BaseTestClass):
            """Tap Test Class."""

            @pytest.fixture
            def config(self) -> SuiteConfig:  # noqa: PLR6301
                return suite_config or SuiteConfig()

            @pytest.fixture
            def resource(self) -> t.Any:  # noqa: ANN401, PLR6301, PT004
                yield  # noqa: PT022

            @pytest.fixture(scope="class")
            def runner(self) -> TapTestRunner | TargetTestRunner:  # noqa: PLR6301
                # Populate runner class with cached records for use in tests
                test_runner.sync_all()
                return test_runner

        return TapTestClass

    def _annotate_test_class(
        self,
        empty_test_class: type[BaseTestClass],
        test_suites: list,
        test_runner: TapTestRunner,
    ) -> type[BaseTestClass]:
        """Annotate test class with test methods.

        Args:
            empty_test_class: Empty test class to be annotated.
            test_suites: List of test suites to include in the test class.
            test_runner: Test runner to be used in the test class.

        Returns:
            An annotated test class.
        """
        for suite in test_suites:
            if suite.kind == "tap":
                self._with_tap_tests(empty_test_class, suite)

            if suite.kind in {"tap_stream", "tap_stream_attribute"}:
                streams = list(test_runner.new_tap().streams.values())

                if suite.kind == "tap_stream":
                    self._with_stream_tests(empty_test_class, suite, streams)

                if suite.kind == "tap_stream_attribute":
                    self._with_stream_attribute_tests(empty_test_class, suite, streams)

        return empty_test_class

    def _with_tap_tests(  # noqa: PLR6301
        self,
        empty_test_class: type[BaseTestClass],
        suite: TestSuite[TapTestTemplate],
    ) -> None:
        for test_class in suite.tests:
            test = test_class()
            test_name = f"test_{suite.kind}_{test.name}"
            setattr(empty_test_class, test_name, test.run)

    def _with_stream_tests(  # noqa: PLR6301
        self,
        empty_test_class: type[BaseTestClass],
        suite: TestSuite[StreamTestTemplate],
        streams: list[Stream],
    ) -> None:
        params = [
            {
                "stream": stream,
            }
            for stream in streams
        ]
        param_ids = [stream.name for stream in streams]

        for test_class in suite.tests:
            test = test_class()
            test_name = f"test_{suite.kind}_{test.name}"
            setattr(
                empty_test_class,
                test_name,
                test.run,
            )
            empty_test_class.params[test_name] = params
            empty_test_class.param_ids[test_name] = param_ids

    def _with_stream_attribute_tests(  # noqa: PLR6301
        self,
        empty_test_class: type[BaseTestClass],
        suite: TestSuite[AttributeTestTemplate],
        streams: list[Stream],
    ) -> None:
        for test_class in suite.tests:
            test = test_class()
            test_name = f"test_{suite.kind}_{test.name}"
            test_params = []
            test_ids: list[str] = []
            for stream in streams:
                final_schema = stream.stream_maps[-1].transformed_schema["properties"]
                test_params.extend(
                    [
                        {
                            "stream": stream,
                            "attribute_name": prop_name,
                        }
                        for prop_name, prop_schema in final_schema.items()
                        if test_class.evaluate(
                            stream=stream,
                            property_name=prop_name,
                            property_schema=prop_schema,
                        )
                    ],
                )
                test_ids.extend(
                    [
                        f"{stream.name}.{prop_name}"
                        for prop_name, prop_schema in final_schema.items()
                        if test_class.evaluate(
                            stream=stream,
                            property_name=prop_name,
                            property_schema=prop_schema,
                        )
                    ],
                )

            if test_params:
                setattr(
                    empty_test_class,
                    test_name,
                    test.run,
                )
                empty_test_class.params[test_name] = test_params
                empty_test_class.param_ids[test_name] = test_ids


class TargetTestClassFactory:
    """Factory for Target Test Classes."""

    def __init__(self, target_class: type[Target], *, config: dict | None = None):
        """Initialize TargetTestClassFactory.

        Args:
            target_class: Target class to be tested.
            config: Config to be used when instantiating tests.
        """
        self.target_class = target_class
        self.config = config

    def new_test_class(
        self,
        *,
        custom_suites: list | None = None,
        suite_config: SuiteConfig | None = None,
        include_target_tests: bool = True,
        **kwargs: t.Any,
    ) -> type[BaseTestClass]:
        """Get a new Target test class.

        Args:
            custom_suites: List of custom test suites to include in the test class.
            suite_config: SuiteConfig instance to be used when instantiating tests.
            include_target_tests: Whether to include target tests in the test class.
            kwargs: Keyword arguments to be passed to the Target on run.

        Returns:
            A new Target test class.
        """
        # compile test suites
        suites = custom_suites or []
        if include_target_tests:
            suites.append(target_tests)

        # set default values
        if "parse_env_config" not in kwargs:
            kwargs["parse_env_config"] = True

        empty_test_class = self._get_empty_test_class(
            target_class=self.target_class,
            config=self.config,
            suite_config=suite_config,
            **kwargs,
        )
        return self._annotate_test_class(
            empty_test_class=empty_test_class,
            test_suites=suites,
        )

    def _get_empty_test_class(  # noqa: PLR6301
        self,
        target_class: type[Target],
        suite_config: SuiteConfig | None,
        config: dict | None = None,
        **kwargs: t.Any,
    ) -> type[BaseTestClass]:
        """Get an empty test class.

        Args:
            target_class: Target class to be tested.
            suite_config: SuiteConfig instance to be used when instantiating tests.
            config: Config to be used when instantiating tests.
            kwargs: Keyword arguments to be passed to the Target on run.

        Returns:
            An empty test class.
        """

        class TargetTestClass(BaseTestClass):
            """Target Test Class."""

            @pytest.fixture
            def config(self) -> SuiteConfig:  # noqa: PLR6301
                return suite_config or SuiteConfig()

            @pytest.fixture
            def resource(self) -> t.Any:  # noqa: ANN401, PLR6301, PT004
                yield  # noqa: PT022

            @pytest.fixture
            def runner(self) -> TargetTestRunner:  # noqa: PLR6301
                # Instantiate new runner class and populate records for use in tests
                return TargetTestRunner(
                    target_class=target_class,
                    config=config,
                    suite_config=suite_config,
                    **kwargs,
                )

        return TargetTestClass

    def _annotate_test_class(  # noqa: PLR6301
        self,
        empty_test_class: type[BaseTestClass],
        test_suites: list,
    ) -> type[BaseTestClass]:
        """Annotate test class with test methods.

        Args:
            empty_test_class: Empty test class to be annotated.
            test_suites: List of test suites to be included in the test class.

        Returns:
            Annotated test class.
        """
        for suite in test_suites:
            if suite.kind == "target":
                for test_class in suite.tests:
                    test = test_class()
                    test_name = f"test_{suite.kind}_{test.name}"
                    setattr(empty_test_class, test_name, test.run)

        return empty_test_class


def get_tap_test_class(
    tap_class: type[Tap],
    *,
    config: dict | None = None,
    include_tap_tests: bool = True,
    include_stream_tests: bool = True,
    include_stream_attribute_tests: bool = True,
    custom_suites: list | None = None,
    suite_config: SuiteConfig | None = None,
    **kwargs: t.Any,
) -> type[BaseTestClass]:
    """Get Tap Test Class.

    Args:
        tap_class: Meltano Singer SDK Tap class to test.
        config: Config dict to use for testing.
        include_tap_tests: Include tap tests.
        include_stream_tests: Include Tap stream tests.
        include_stream_attribute_tests: Include Tap stream attribute tests.
        custom_suites: Custom test suites to add to standard tests.
        suite_config: SuiteConfig instance to pass to tests.
        kwargs: Keyword arguments to pass to the TapRunner.

    Returns:
        A test class usable by pytest.
    """
    factory = TapTestClassFactory(
        tap_class=tap_class,
        config=config,
    )
    return factory.new_test_class(
        custom_suites=custom_suites,
        suite_config=suite_config,
        include_tap_tests=include_tap_tests,
        include_stream_tests=include_stream_tests,
        include_stream_attribute_tests=include_stream_attribute_tests,
        **kwargs,
    )


def get_target_test_class(
    target_class: type[Target],
    *,
    config: dict | None = None,
    custom_suites: list | None = None,
    suite_config: SuiteConfig | None = None,
    include_target_tests: bool = True,
    **kwargs: t.Any,
) -> type[BaseTestClass]:
    """Get Target Test Class.

    Args:
        target_class: Meltano Singer SDK Target class to test.
        config: Config dict to use for testing.
        custom_suites: Custom test suites to add to standard tests.
        suite_config: SuiteConfig instance to pass to tests.
        include_target_tests: Include standard target tests.
        kwargs: Keyword arguments to pass to the TapRunner.

    Returns:
        A test class usable by pytest.
    """
    factory = TargetTestClassFactory(
        target_class=target_class,
        config=config,
    )
    return factory.new_test_class(
        custom_suites=custom_suites,
        suite_config=suite_config,
        include_target_tests=include_target_tests,
        **kwargs,
    )



"""About information for a plugin."""

from __future__ import annotations

import abc
import dataclasses
import json
import typing as t
from collections import OrderedDict
from textwrap import dedent

from packaging.specifiers import SpecifierSet
from packaging.version import Version

if t.TYPE_CHECKING:
    from singer_sdk.helpers.capabilities import CapabilitiesEnum

__all__ = [
    "AboutFormatter",
    "AboutInfo",
    "JSONFormatter",
    "MarkdownFormatter",
]

# Keep these in sync with the supported Python versions in pyproject.toml
_PY_MIN_VERSION = 8
_PY_MAX_VERSION = 12


def _get_min_version(specifiers: SpecifierSet) -> int:
    min_version: list[int] = []
    for specifier in specifiers:
        if specifier.operator == ">=":
            min_version.append(Version(specifier.version).minor)
        if specifier.operator == ">":
            min_version.append(Version(specifier.version).minor + 1)
    return min(min_version, default=_PY_MIN_VERSION)


def _get_max_version(specifiers: SpecifierSet) -> int:
    max_version: list[int] = []
    for specifier in specifiers:
        if specifier.operator == "<=":
            max_version.append(Version(specifier.version).minor)
        if specifier.operator == "<":
            max_version.append(Version(specifier.version).minor - 1)
    return max(max_version, default=_PY_MAX_VERSION)


def get_supported_pythons(requires_python: str) -> t.Generator[str, None, None]:
    specifiers = SpecifierSet(requires_python)
    min_version = _get_min_version(specifiers)
    max_version = _get_max_version(specifiers)

    yield from specifiers.filter(f"3.{v}" for v in range(min_version, max_version + 1))


@dataclasses.dataclass
class AboutInfo:
    """About information for a plugin."""

    name: str
    description: str | None
    version: str
    sdk_version: str
    supported_python_versions: list[str] | None

    capabilities: list[CapabilitiesEnum]
    settings: dict
    env_var_prefix: str


class AboutFormatter(abc.ABC):
    """Abstract base class for about formatters."""

    formats: t.ClassVar[dict[str, type[AboutFormatter]]] = {}
    format_name: str

    def __init_subclass__(cls, format_name: str) -> None:
        """Initialize subclass.

        Args:
            format_name: Name of the format.
        """
        cls.formats[format_name] = cls
        super().__init_subclass__()

    @classmethod
    def get_formatter(cls, name: str) -> AboutFormatter:
        """Get a formatter by name.

        Args:
            name: Name of the formatter.

        Returns:
            A formatter.
        """
        return cls.formats[name]()

    @abc.abstractmethod
    def format_about(self, about_info: AboutInfo) -> str:
        """Render about information.

        Args:
            about_info: About information.
        """
        ...


class TextFormatter(AboutFormatter, format_name="text"):
    """About formatter for text output."""

    def format_about(self, about_info: AboutInfo) -> str:  # noqa: PLR6301
        """Render about information.

        Args:
            about_info: About information.

        Returns:
            A formatted string.
        """
        output = dedent(
            f"""\
            Name: {about_info.name}
            Description: {about_info.description}
            Version: {about_info.version}
            SDK Version: {about_info.sdk_version}"""
        )

        if about_info.supported_python_versions:
            output += "\nSupported Python Versions:\n"
            output += "\n".join(
                [f"  - {v}" for v in about_info.supported_python_versions]
            )

        output += "\nCapabilities:\n"
        output += "\n".join([f"  - {c}" for c in about_info.capabilities])

        output += "\nSettings:\n"
        for setting, schema in about_info.settings.get("properties", {}).items():
            env_var = about_info.env_var_prefix + setting.upper().replace("-", "_")
            json_type = schema.get("type")
            output += f"  - Name: {setting}\n"
            output += f"    Type: {json_type}\n"
            output += f"    Environment Variable: {env_var}\n"

        return output


class JSONFormatter(AboutFormatter, format_name="json"):
    """About formatter for JSON output."""

    def __init__(self) -> None:
        """Initialize a JSONAboutFormatter."""
        self.indent = 2
        self.default = str

    def format_about(self, about_info: AboutInfo) -> str:
        """Render about information.

        Args:
            about_info: About information.

        Returns:
            A formatted string.
        """
        data = OrderedDict(
            [
                ("name", about_info.name),
                ("description", about_info.description),
                ("version", about_info.version),
                ("sdk_version", about_info.sdk_version),
                ("supported_python_versions", about_info.supported_python_versions),
                ("capabilities", [c.value for c in about_info.capabilities]),
                ("settings", about_info.settings),
            ],
        )
        return json.dumps(data, indent=self.indent, default=self.default)


class MarkdownFormatter(AboutFormatter, format_name="markdown"):
    """About formatter for Markdown output."""

    def _generate_property_row(
        self,
        name: str,
        schema: dict[str, t.Any],
        *,
        required: bool,
        parent_name: str | None = None,
    ) -> t.Generator[str, None, None]:
        """Generate a row for this property and for nested properties, if any.

        Args:
            name: The property name.
            schema: The property schema.
            required: Whether the property is required to be present.
            parent_name: The parent property name, if any.

        Yields:
            One markdown table row for the setting, and one for each sub-property.
        """
        setting_name = f"{parent_name}.{name}" if parent_name else name
        md_description = schema.get("description", "").replace("\n", "<BR/>")
        yield (
            f"| {setting_name} "
            f"| {'True' if required else 'False':8} "
            f"| {schema.get('default', 'None'):7} "
            f"| {md_description:11} |"
        )
        if "properties" in schema:
            yield from self._generate_property_rows(schema, parent_name=setting_name)

    def _generate_property_rows(
        self,
        schema: dict[str, t.Any],
        *,
        parent_name: str = "",
    ) -> t.Generator[str, None, None]:
        """Generate a row for each property in the schema.

        Args:
            schema: A JSON object schema.
            parent_name: The parent property name, if any.

        Yields:
            One markdown table row for each property.
        """
        required_settings = schema.get("required", [])
        for name, sub_schema in schema.get("properties", {}).items():
            yield from self._generate_property_row(
                name,
                sub_schema,
                required=name in required_settings,
                parent_name=parent_name,
            )

    def format_about(self, about_info: AboutInfo) -> str:
        """Render about information.

        Args:
            about_info: About information.

        Returns:
            A formatted string.
        """
        # Header
        output = dedent(f"""\
            # `{about_info.name}`\n
            {about_info.description}\n
            Built with the [Meltano Singer SDK](https://sdk.meltano.com).\n
        """)

        # Process capabilities
        output += "## Capabilities\n\n"
        output += "\n".join([f"* `{v}`" for v in about_info.capabilities])
        output += "\n\n"

        # Process Supported Python Versions
        if about_info.supported_python_versions:
            output += "## Supported Python Versions\n\n"
            output += "\n".join(
                [f"* {v}" for v in about_info.supported_python_versions],
            )
            output += "\n\n"

        # Process settings
        output += "## Settings\n\n"
        output += (
            "| Setting | Required | Default | Description |\n"
            "|:--------|:--------:|:-------:|:------------|\n"
        )
        output += "\n".join(self._generate_property_rows(about_info.settings))
        output += (
            "\n\n"
            + "\n".join(
                [
                    "A full list of supported settings and capabilities "
                    f"is available by running: `{about_info.name} --about`",
                ],
            )
            + "\n"
        )

        return output



"""Abstract base class for stream mapper plugins."""

from __future__ import annotations

import abc
import typing as t

import click

from singer_sdk.helpers._classproperty import classproperty
from singer_sdk.helpers.capabilities import CapabilitiesEnum, PluginCapabilities
from singer_sdk.io_base import SingerReader, SingerWriter
from singer_sdk.plugin_base import PluginBase

if t.TYPE_CHECKING:
    import singer_sdk._singerlib as singer


class InlineMapper(PluginBase, SingerReader, SingerWriter, metaclass=abc.ABCMeta):
    """Abstract base class for inline mappers."""

    @classproperty
    def capabilities(self) -> list[CapabilitiesEnum]:  # noqa: PLR6301
        """Get capabilities.

        Returns:
            A list of plugin capabilities.
        """
        return [
            PluginCapabilities.STREAM_MAPS,
        ]

    def _write_messages(self, messages: t.Iterable[singer.Message]) -> None:
        for message in messages:
            self.write_message(message)

    def _process_schema_message(self, message_dict: dict) -> None:
        self._write_messages(self.map_schema_message(message_dict))

    def _process_record_message(self, message_dict: dict) -> None:
        self._write_messages(self.map_record_message(message_dict))

    def _process_state_message(self, message_dict: dict) -> None:
        self._write_messages(self.map_state_message(message_dict))

    def _process_activate_version_message(self, message_dict: dict) -> None:
        self._write_messages(self.map_activate_version_message(message_dict))

    def _process_batch_message(self, message_dict: dict) -> None:
        self._write_messages(self.map_batch_message(message_dict))

    @abc.abstractmethod
    def map_schema_message(self, message_dict: dict) -> t.Iterable[singer.Message]:
        """Map a schema message to zero or more new messages.

        Args:
            message_dict: A SCHEMA message JSON dictionary.
        """
        ...

    @abc.abstractmethod
    def map_record_message(self, message_dict: dict) -> t.Iterable[singer.Message]:
        """Map a record message to zero or more new messages.

        Args:
            message_dict: A RECORD message JSON dictionary.
        """
        ...

    @abc.abstractmethod
    def map_state_message(self, message_dict: dict) -> t.Iterable[singer.Message]:
        """Map a state message to zero or more new messages.

        Args:
            message_dict: A STATE message JSON dictionary.
        """
        ...

    @abc.abstractmethod
    def map_activate_version_message(
        self,
        message_dict: dict,
    ) -> t.Iterable[singer.Message]:
        """Map a version message to zero or more new messages.

        Args:
            message_dict: An ACTIVATE_VERSION message JSON dictionary.
        """
        ...

    def map_batch_message(  # noqa: PLR6301
        self,
        message_dict: dict,  # noqa: ARG002
    ) -> t.Iterable[singer.Message]:
        """Map a batch message to zero or more new messages.

        Args:
            message_dict: A BATCH message JSON dictionary.

        Raises:
            NotImplementedError: if not implemented by subclass.
        """
        msg = "BATCH messages are not supported by mappers."
        raise NotImplementedError(msg)

    # CLI handler

    @classmethod
    def invoke(  # type: ignore[override]
        cls: type[InlineMapper],
        *,
        about: bool = False,
        about_format: str | None = None,
        config: tuple[str, ...] = (),
        file_input: t.IO[str] | None = None,
    ) -> None:
        """Invoke the mapper.

        Args:
            about: Display package metadata and settings.
            about_format: Specify output style for `--about`.
            config: Configuration file location or 'ENV' to use environment
                variables. Accepts multiple inputs as a tuple.
            file_input: Optional file to read input from.
        """
        super().invoke(about=about, about_format=about_format)
        cls.print_version(print_fn=cls.logger.info)
        config_files, parse_env_config = cls.config_from_cli_args(*config)

        mapper = cls(
            config=config_files,  # type: ignore[arg-type]
            validate_config=True,
            parse_env_config=parse_env_config,
        )
        mapper.listen(file_input)

    @classmethod
    def get_singer_command(cls: type[InlineMapper]) -> click.Command:
        """Execute standard CLI handler for inline mappers.

        Returns:
            A click.Command object.
        """
        command = super().get_singer_command()
        command.help = "Execute the Singer mapper."
        command.params.extend(
            [
                click.Option(
                    ["--input", "file_input"],
                    help="A path to read messages from instead of from standard in.",
                    type=click.File("r"),
                ),
            ],
        )

        return command



"""Sink classes load data to a target."""

from __future__ import annotations

import abc
import copy
import datetime
import importlib.util
import time
import typing as t
from functools import cached_property
from gzip import GzipFile
from gzip import open as gzip_open
from types import MappingProxyType

import jsonschema
from typing_extensions import override

from singer_sdk._singerlib.json import deserialize_json
from singer_sdk.exceptions import (
    InvalidJSONSchema,
    InvalidRecord,
    MissingKeyPropertiesError,
)
from singer_sdk.helpers._batch import (
    BaseBatchFileEncoding,
    BatchConfig,
    BatchFileFormat,
    StorageTarget,
)
from singer_sdk.helpers._compat import (
    date_fromisoformat,
    datetime_fromisoformat,
    time_fromisoformat,
)
from singer_sdk.helpers._typing import (
    DatetimeErrorTreatmentEnum,
    get_datelike_property_type,
    handle_invalid_timestamp_in_record,
)

if t.TYPE_CHECKING:
    from logging import Logger

    from singer_sdk.target_base import Target


class BaseJSONSchemaValidator(abc.ABC):
    """Abstract base class for JSONSchema validator."""

    def __init__(self, schema: dict[str, t.Any]) -> None:
        """Initialize the record validator.

        Args:
            schema: Schema of the stream to sink.
        """
        self.schema = schema

    @abc.abstractmethod
    def validate(self, record: dict[str, t.Any]) -> None:
        """Validate a record message.

        This method MUST raise an ``InvalidRecord`` exception if the record is invalid.

        Args:
            record: Record message to validate.
        """


class JSONSchemaValidator(BaseJSONSchemaValidator):
    """Validate records using the ``fastjsonschema`` library."""

    def __init__(
        self,
        schema: dict,
        *,
        validate_formats: bool = False,
        format_checker: jsonschema.FormatChecker | None = None,
    ):
        """Initialize the validator.

        Args:
            schema: Schema of the stream to sink.
            validate_formats: Whether JSON string formats (e.g. ``date-time``) should
                be validated.
            format_checker: User-defined format checker.

        Raises:
            InvalidJSONSchema: If the schema provided from tap or mapper is invalid.
        """
        jsonschema_validator = jsonschema.Draft7Validator

        super().__init__(schema)
        if validate_formats:
            format_checker = format_checker or jsonschema_validator.FORMAT_CHECKER
        else:
            format_checker = jsonschema.FormatChecker(formats=())

        try:
            jsonschema_validator.check_schema(schema)
        except jsonschema.SchemaError as e:
            error_message = f"Schema Validation Error: {e}"
            raise InvalidJSONSchema(error_message) from e

        self.validator = jsonschema_validator(
            schema=schema,
            format_checker=format_checker,
        )

    @override
    def validate(self, record: dict):  # noqa: ANN201
        """Validate a record message.

        Args:
            record: Record message to validate.

        Raises:
            InvalidRecord: If the record is invalid.
        """
        try:
            self.validator.validate(record)
        except jsonschema.ValidationError as e:
            raise InvalidRecord(e.message, record) from e


class Sink(metaclass=abc.ABCMeta):  # noqa: PLR0904
    """Abstract base class for target sinks."""

    # max timestamp/datetime supported, used to reset invalid dates

    logger: Logger

    MAX_SIZE_DEFAULT = 10000

    validate_field_string_format = False
    """Enable JSON schema format validation, for example `date-time` string fields."""

    fail_on_record_validation_exception: bool = True
    """Interrupt the target execution when a record fails schema validation."""

    def __init__(
        self,
        target: Target,
        stream_name: str,
        schema: dict,
        key_properties: t.Sequence[str] | None,
    ) -> None:
        """Initialize target sink.

        Args:
            target: Target instance.
            stream_name: Name of the stream to sink.
            schema: Schema of the stream to sink.
            key_properties: Primary key of the stream to sink.
        """
        self.logger = target.logger.getChild(stream_name)
        self.sync_started_at = target.initialized_at
        self._config = dict(target.config)
        self._pending_batch: dict | None = None
        self.stream_name = stream_name
        self.logger.info(
            "Initializing target sink for stream '%s'...",
            stream_name,
        )
        self.original_schema = copy.deepcopy(schema)
        self.schema = schema
        if self.include_sdc_metadata_properties:
            self._add_sdc_metadata_to_schema()
        else:
            self._remove_sdc_metadata_from_schema()
        self.records_to_drain: list[dict] | t.Any = []
        self._context_draining: dict | None = None
        self.latest_state: dict | None = None
        self._draining_state: dict | None = None
        self.drained_state: dict | None = None
        self._key_properties = key_properties or []

        # Tally counters
        self._total_records_written: int = 0
        self._total_dupe_records_merged: int = 0
        self._total_records_read: int = 0
        self._batch_records_read: int = 0
        self._batch_dupe_records_merged: int = 0

        # Batch full markers
        self._batch_size_rows: int | None = target.config.get(
            "batch_size_rows",
        )

        self._validator: BaseJSONSchemaValidator | None = self.get_validator()

    @cached_property
    def validate_schema(self) -> bool:
        """Enable JSON schema record validation.

        Returns:
            True if JSON schema validation is enabled.
        """
        return self.config.get("validate_records", True)

    def get_validator(self) -> BaseJSONSchemaValidator | None:
        """Get a record validator for this sink.

        Override this method to use a custom format validator, or disable record
        validation by returning `None`.

        Returns:
            An instance of a subclass of ``BaseJSONSchemaValidator``.

        Example implementation using the `fastjsonschema`_ library:

        .. code-block:: python

           import fastjsonschema


           class FastJSONSchemaValidator(BaseJSONSchemaValidator):
               def __init__(self, schema: dict[str, t.Any]) -> None:
                   super().__init__(schema)
                   try:
                       self.validator = fastjsonschema.compile(self.schema)
                   except fastjsonschema.JsonSchemaDefinitionException as e:
                       error_message = "Schema Validation Error"
                       raise InvalidJSONSchema(error_message) from e

               def validate(self, record: dict):
                   try:
                       self.validator(record)
                   except fastjsonschema.JsonSchemaValueException as e:
                       error_message = f"Record Message Validation Error: {e.message}"
                       raise InvalidRecord(error_message, record) from e

        .. _fastjsonschema: https://pypi.org/project/fastjsonschema/
        """
        if self.validate_schema:
            return JSONSchemaValidator(
                self.schema,
                validate_formats=self.validate_field_string_format,
            )
        return None

    def _get_context(self, record: dict) -> dict:  # noqa: ARG002, PLR6301
        """Return an empty dictionary by default.

        NOTE: Future versions of the SDK may expand the available context attributes.

        Args:
            record: Individual record in the stream.

        Returns:
            TODO
        """
        return {}

    # Size properties

    @property
    def current_size(self) -> int:
        """Get current batch size.

        Returns:
            The number of records to drain.
        """
        return self._batch_records_read

    @property
    def is_full(self) -> bool:
        """Check against the batch size limit.

        Returns:
            True if the sink needs to be drained.
        """
        return self.current_size >= self.max_size

    @property
    def batch_size_rows(self) -> int | None:
        """The maximum number of rows a batch can accumulate before being processed.

        Returns:
            The max number of rows or None if not set.
        """
        return self._batch_size_rows

    @property
    def max_size(self) -> int:
        """Get max batch size.

        Returns:
            Max number of records to batch before `is_full=True`

        .. versionchanged:: 0.36.0
           This property now takes into account the
           :attr:`~singer_sdk.Sink.batch_size_rows` attribute and the corresponding
           ``batch_size_rows`` target setting.
        """
        return (
            self.batch_size_rows
            if self.batch_size_rows is not None
            else self.MAX_SIZE_DEFAULT
        )

    # Tally methods

    @t.final
    def tally_record_read(self, count: int = 1) -> None:
        """Increment the records read tally.

        This method is called automatically by the SDK when records are read.

        Args:
            count: Number to increase record count by.
        """
        self._total_records_read += count
        self._batch_records_read += count

    @t.final
    def tally_record_written(self, count: int = 1) -> None:
        """Increment the records written tally.

        This method is called automatically by the SDK after
        :meth:`~singer_sdk.Sink.process_record()`
        or :meth:`~singer_sdk.Sink.process_batch()`.

        Args:
            count: Number to increase record count by.
        """
        self._total_records_written += count

    @t.final
    def tally_duplicate_merged(self, count: int = 1) -> None:
        """Increment the records merged tally.

        This method should be called directly by the Target implementation.

        Args:
            count: Number to increase record count by.
        """
        self._total_dupe_records_merged += count
        self._batch_dupe_records_merged += count

    # Properties

    @property
    def config(self) -> t.Mapping[str, t.Any]:
        """Get plugin configuration.

        Returns:
            A frozen (read-only) config dictionary map.
        """
        return MappingProxyType(self._config)

    @property
    def batch_config(self) -> BatchConfig | None:
        """Get batch configuration.

        Returns:
            A frozen (read-only) config dictionary map.
        """
        raw = self.config.get("batch_config")
        return BatchConfig.from_dict(raw) if raw else None

    @property
    def include_sdc_metadata_properties(self) -> bool:
        """Check if metadata columns should be added.

        Returns:
            True if metadata columns should be added.
        """
        return self.config.get("add_record_metadata", False)

    @property
    def datetime_error_treatment(self) -> DatetimeErrorTreatmentEnum:
        """Return a treatment to use for datetime parse errors: ERROR. MAX, or NULL.

        Returns:
            TODO
        """
        return DatetimeErrorTreatmentEnum.ERROR

    @property
    def key_properties(self) -> t.Sequence[str]:
        """Return key properties.

        Override this method to return a list of key properties in a format that is
        compatible with the target.

        Returns:
            A list of stream key properties.
        """
        return self._key_properties

    # Record processing

    def _add_sdc_metadata_to_record(
        self,
        record: dict,
        message: dict,
        context: dict,
    ) -> None:
        """Populate metadata _sdc columns from incoming record message.

        Record metadata specs documented at:
        https://sdk.meltano.com/en/latest/implementation/record_metadata.html

        Args:
            record: Individual record in the stream.
            message: The record message.
            context: Stream partition or context dictionary.
        """
        record["_sdc_extracted_at"] = message.get("time_extracted")
        record["_sdc_received_at"] = datetime.datetime.now(
            tz=datetime.timezone.utc,
        ).isoformat()
        record["_sdc_batched_at"] = (
            context.get("batch_start_time")
            or datetime.datetime.now(tz=datetime.timezone.utc)
        ).isoformat()
        record["_sdc_deleted_at"] = record.get("_sdc_deleted_at")
        record["_sdc_sequence"] = int(round(time.time() * 1000))
        record["_sdc_table_version"] = message.get("version")
        record["_sdc_sync_started_at"] = self.sync_started_at

    def _add_sdc_metadata_to_schema(self) -> None:
        """Add _sdc metadata columns.

        Record metadata specs documented at:
        https://sdk.meltano.com/en/latest/implementation/record_metadata.html
        """
        properties_dict = self.schema["properties"]
        for col in (
            "_sdc_extracted_at",
            "_sdc_received_at",
            "_sdc_batched_at",
            "_sdc_deleted_at",
        ):
            properties_dict[col] = {
                "type": ["null", "string"],
                "format": "date-time",
            }
        for col in ("_sdc_sequence", "_sdc_table_version", "_sdc_sync_started_at"):
            properties_dict[col] = {"type": ["null", "integer"]}

    def _remove_sdc_metadata_from_schema(self) -> None:
        """Remove _sdc metadata columns.

        Record metadata specs documented at:
        https://sdk.meltano.com/en/latest/implementation/record_metadata.html
        """
        properties_dict = self.schema["properties"]
        for col in (
            "_sdc_extracted_at",
            "_sdc_received_at",
            "_sdc_batched_at",
            "_sdc_deleted_at",
            "_sdc_sequence",
            "_sdc_table_version",
            "_sdc_sync_started_at",
        ):
            properties_dict.pop(col, None)

    def _remove_sdc_metadata_from_record(self, record: dict) -> None:  # noqa: PLR6301
        """Remove metadata _sdc columns from incoming record message.

        Record metadata specs documented at:
        https://sdk.meltano.com/en/latest/implementation/record_metadata.html

        Args:
            record: Individual record in the stream.
        """
        record.pop("_sdc_extracted_at", None)
        record.pop("_sdc_received_at", None)
        record.pop("_sdc_batched_at", None)
        record.pop("_sdc_deleted_at", None)
        record.pop("_sdc_sequence", None)
        record.pop("_sdc_table_version", None)
        record.pop("_sdc_sync_started_at", None)

    # Record validation

    def _validate_and_parse(self, record: dict) -> dict:
        """Validate or repair the record, parsing to python-native types as needed.

        Args:
            record: Individual record in the stream.

        Returns:
            TODO

        Raises:
            InvalidRecord: If the record is invalid.
        """
        if self._validator is not None:
            # TODO: Check the performance impact of this try/except block. It runs
            # on every record, so it's probably bad and should be moved up the stack.
            try:
                self._validator.validate(record)
            except InvalidRecord:
                self.logger.exception("Record validation failed")
                if self.fail_on_record_validation_exception:
                    raise

        self._parse_timestamps_in_record(
            record=record,
            schema=self.schema,
            treatment=self.datetime_error_treatment,
        )
        return record

    def _singer_validate_message(self, record: dict) -> None:
        """Ensure record conforms to Singer Spec.

        Args:
            record: Record (after parsing, schema validations and transformations).

        Raises:
            MissingKeyPropertiesError: If record is missing one or more key properties.
        """
        if any(key_property not in record for key_property in self._key_properties):
            msg = (
                f"Record is missing one or more key_properties. \n"
                f"Key Properties: {self._key_properties}, "
                f"Record Keys: {list(record.keys())}"
            )
            raise MissingKeyPropertiesError(
                msg,
            )

    def _parse_timestamps_in_record(
        self,
        record: dict,
        schema: dict,
        treatment: DatetimeErrorTreatmentEnum,
    ) -> None:
        """Parse strings to datetime.datetime values, repairing or erroring on failure.

        Attempts to parse every field that is of type date/datetime/time. If its value
        is out of range, repair logic will be driven by the `treatment` input arg:
        MAX, NULL, or ERROR.

        Args:
            record: Individual record in the stream.
            schema: TODO
            treatment: TODO
        """
        for key, value in record.items():
            if key not in schema["properties"]:
                if value is not None:
                    self.logger.warning("No schema for record field '%s'", key)
                continue
            datelike_type = get_datelike_property_type(schema["properties"][key])
            if datelike_type:
                date_val = value
                try:
                    if value is not None:
                        if datelike_type == "time":
                            date_val = time_fromisoformat(date_val)
                        elif datelike_type == "date":
                            date_val = date_fromisoformat(date_val)
                        else:
                            date_val = datetime_fromisoformat(date_val)
                except ValueError as ex:
                    date_val = handle_invalid_timestamp_in_record(
                        record,
                        [key],
                        date_val,
                        datelike_type,
                        ex,
                        treatment,
                        self.logger,
                    )
                record[key] = date_val

    def _after_process_record(self, context: dict) -> None:
        """Perform post-processing and record keeping. Internal hook.

        Args:
            context: Stream partition or context dictionary.
        """
        self.logger.debug("Processed record: %s", context)

    # SDK developer overrides:

    def preprocess_record(self, record: dict, context: dict) -> dict:  # noqa: ARG002, PLR6301
        """Process incoming record and return a modified result.

        Args:
            record: Individual record in the stream.
            context: Stream partition or context dictionary.

        Returns:
            A new, processed record.
        """
        return record

    @abc.abstractmethod
    def process_record(self, record: dict, context: dict) -> None:
        """Load the latest record from the stream.

        Implementations may either load to the `context` dict for staging (the
        default behavior for Batch types), or permanently write out to the target.

        Anything appended to :attr:`singer_sdk.Sink.records_to_drain` will be
        automatically passed to
        :meth:`~singer_sdk.Sink.process_batch()` to be permanently written during the
        process_batch operation.

        If duplicates are merged, these can be tracked via
        :meth:`~singer_sdk.Sink.tally_duplicate_merged()`.

        Args:
            record: Individual record in the stream.
            context: Stream partition or context dictionary.
        """

    def start_drain(self) -> dict:
        """Set and return `self._context_draining`.

        Returns:
            TODO
        """
        self._context_draining = self._pending_batch or {}
        self._pending_batch = None
        return self._context_draining

    @abc.abstractmethod
    def process_batch(self, context: dict) -> None:
        """Process all records per the batch's `context` dictionary.

        If duplicates are merged, these can optionally be tracked via
        `tally_duplicate_merged()`.

        Args:
            context: Stream partition or context dictionary.

        Raises:
            NotImplementedError: If derived class does not override this method.
        """
        msg = "No handling exists for process_batch()."
        raise NotImplementedError(msg)

    def mark_drained(self) -> None:
        """Reset `records_to_drain` and any other tracking."""
        self.drained_state = self._draining_state
        self._draining_state = None
        self._context_draining = None
        if self._batch_records_read:
            self.tally_record_written(
                self._batch_records_read - self._batch_dupe_records_merged,
            )
        self._batch_records_read = 0

    def activate_version(self, new_version: int) -> None:
        """Bump the active version of the target table.

        This method should be overridden by developers if a custom implementation is
        expected.

        Args:
            new_version: The version number to activate.
        """
        _ = new_version
        self.logger.warning(
            "ACTIVATE_VERSION message received but not implemented by this target. "
            "Ignoring.",
        )

    def setup(self) -> None:
        """Perform any setup actions at the beginning of a Stream.

        Setup is executed once per Sink instance, after instantiation. If a Schema
        change is detected, a new Sink is instantiated and this method is called again.
        """
        self.logger.info("Setting up %s", self.stream_name)

    def clean_up(self) -> None:
        """Perform any clean up actions required at end of a stream.

        Implementations should ensure that clean up does not affect resources
        that may be in use from other instances of the same sink. Stream name alone
        should not be relied on, it's recommended to use a uuid as well.
        """
        self.logger.info("Cleaning up %s", self.stream_name)

    def process_batch_files(
        self,
        encoding: BaseBatchFileEncoding,
        files: t.Sequence[str],
    ) -> None:
        """Process a batch file with the given batch context.

        Args:
            encoding: The batch file encoding.
            files: The batch files to process.

        Raises:
            NotImplementedError: If the batch file encoding is not supported.
        """
        file: GzipFile | t.IO
        storage: StorageTarget | None = None

        for path in files:
            head, tail = StorageTarget.split_url(path)

            if self.batch_config:
                storage = self.batch_config.storage
            else:
                storage = StorageTarget.from_url(head)

            if encoding.format == BatchFileFormat.JSONL:
                with storage.fs(create=False) as batch_fs, batch_fs.open(
                    tail,
                    mode="rb",
                ) as file:
                    context_file = (
                        gzip_open(file) if encoding.compression == "gzip" else file
                    )
                    context = {
                        "records": [deserialize_json(line) for line in context_file]  # type: ignore[attr-defined]
                    }
                    self.process_batch(context)
            elif (
                importlib.util.find_spec("pyarrow")
                and encoding.format == BatchFileFormat.PARQUET
            ):
                import pyarrow.parquet as pq  # noqa: PLC0415

                with storage.fs(create=False) as batch_fs, batch_fs.open(
                    tail,
                    mode="rb",
                ) as file:
                    context_file = file
                    table = pq.read_table(context_file)
                    context = {"records": table.to_pylist()}
                    self.process_batch(context)
            else:
                msg = f"Unsupported batch encoding format: {encoding.format}"
                raise NotImplementedError(msg)



"""Singer metrics logging."""

from __future__ import annotations

import abc
import enum
import json
import logging
import logging.config
import os
import typing as t
from dataclasses import dataclass, field
from pathlib import Path
from time import time

import yaml

from singer_sdk.helpers._resources import get_package_files

if t.TYPE_CHECKING:
    from types import TracebackType

    from singer_sdk.helpers import types
    from singer_sdk.helpers._compat import Traversable


DEFAULT_LOG_INTERVAL = 60.0
METRICS_LOGGER_NAME = __name__
METRICS_LOG_LEVEL_SETTING = "metrics_log_level"

_TVal = t.TypeVar("_TVal")


class Status(str, enum.Enum):
    """Constants for commonly used status values."""

    SUCCEEDED = "succeeded"
    FAILED = "failed"


class Tag(str, enum.Enum):
    """Constants for commonly used tags."""

    STREAM = "stream"
    CONTEXT = "context"
    ENDPOINT = "endpoint"
    JOB_TYPE = "job_type"
    HTTP_STATUS_CODE = "http_status_code"
    STATUS = "status"


class Metric(str, enum.Enum):
    """Common metric types."""

    RECORD_COUNT = "record_count"
    BATCH_COUNT = "batch_count"
    HTTP_REQUEST_DURATION = "http_request_duration"
    HTTP_REQUEST_COUNT = "http_request_count"
    JOB_DURATION = "job_duration"
    SYNC_DURATION = "sync_duration"


@dataclass
class Point(t.Generic[_TVal]):
    """An individual metric measurement."""

    metric_type: str
    metric: Metric
    value: _TVal
    tags: dict[str, t.Any] = field(default_factory=dict)

    def __str__(self) -> str:
        """Get string representation of this measurement.

        Returns:
            A string representation of this measurement.
        """
        return self.to_json()

    def to_json(self) -> str:
        """Convert this measure to a JSON object.

        Returns:
            A JSON object.
        """
        return json.dumps(
            {
                "type": self.metric_type,
                "metric": self.metric.value,
                "value": self.value,
                "tags": self.tags,
            },
            default=str,
        )


def log(logger: logging.Logger, point: Point) -> None:
    """Log a measurement.

    Args:
        logger: An logger instance.
        point: A measurement.
    """
    logger.info("METRIC: %s", point)


class Meter(metaclass=abc.ABCMeta):
    """Base class for all meters."""

    def __init__(self, metric: Metric, tags: dict | None = None) -> None:
        """Initialize a meter.

        Args:
            metric: The metric type.
            tags: Tags to add to the measurement.
        """
        self.metric = metric
        self.tags = tags or {}
        self.logger = get_metrics_logger()

    @property
    def context(self) -> types.Context | None:
        """Get the context for this meter.

        Returns:
            A context dictionary.
        """
        return self.tags.get(Tag.CONTEXT)

    @context.setter
    def context(self, value: types.Context | None) -> None:
        """Set the context for this meter.

        Args:
            value: A context dictionary.
        """
        if value is None:
            self.tags.pop(Tag.CONTEXT, None)
        else:
            self.tags[Tag.CONTEXT] = value

    @abc.abstractmethod
    def __enter__(self) -> Meter:
        """Enter the meter context."""
        ...

    @abc.abstractmethod
    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the meter context.

        Args:
            exc_type: The exception type.
            exc_val: The exception value.
            exc_tb: The exception traceback.
        """
        ...


class Counter(Meter):
    """A meter for counting things."""

    def __init__(
        self,
        metric: Metric,
        tags: dict | None = None,
        log_interval: float = DEFAULT_LOG_INTERVAL,
    ) -> None:
        """Initialize a counter.

        Args:
            metric: The metric type.
            tags: Tags to add to the measurement.
            log_interval: The interval at which to log the count.
        """
        super().__init__(metric, tags)
        self.value = 0
        self.log_interval = log_interval
        self.last_log_time = time()

    def __enter__(self) -> Counter:
        """Enter the counter context.

        Returns:
            The counter instance.
        """
        self.last_log_time = time()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the counter context.

        Args:
            exc_type: The exception type.
            exc_val: The exception value.
            exc_tb: The exception traceback.
        """
        self._pop()

    def _pop(self) -> None:
        """Log and reset the counter."""
        log(self.logger, Point("counter", self.metric, self.value, self.tags))
        self.value = 0
        self.last_log_time = time()

    def increment(self, value: int = 1) -> None:
        """Increment the counter.

        Args:
            value: The value to increment by.
        """
        self.value += value
        if self._ready_to_log():
            self._pop()

    def _ready_to_log(self) -> bool:
        """Check if the counter is ready to log.

        Returns:
            True if the counter is ready to log.
        """
        return time() - self.last_log_time > self.log_interval


class Timer(Meter):
    """A meter for timing things."""

    def __init__(self, metric: Metric, tags: dict | None = None) -> None:
        """Initialize a timer.

        Args:
            metric: The metric type.
            tags: Tags to add to the measurement.
        """
        super().__init__(metric, tags)
        self.start_time = time()

    def __enter__(self) -> Timer:
        """Enter the timer context.

        Returns:
            The timer instance.
        """
        self.start_time = time()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        """Exit the timer context.

        Args:
            exc_type: The exception type.
            exc_val: The exception value.
            exc_tb: The exception traceback.
        """
        if Tag.STATUS not in self.tags:
            self.tags[Tag.STATUS] = (
                Status.SUCCEEDED if exc_type is None else Status.FAILED
            )
        log(self.logger, Point("timer", self.metric, self.elapsed(), self.tags))

    def elapsed(self) -> float:
        """Get the elapsed time.

        Returns:
            The elapsed time.
        """
        return time() - self.start_time


def get_metrics_logger() -> logging.Logger:
    """Get a logger for emitting metrics.

    Returns:
        A logger that can be used to emit metrics.
    """
    return logging.getLogger(METRICS_LOGGER_NAME)


def record_counter(
    stream: str,
    endpoint: str | None = None,
    log_interval: float = DEFAULT_LOG_INTERVAL,
    **tags: t.Any,
) -> Counter:
    """Use for counting records retrieved from the source.

    with record_counter("my_stream", endpoint="/users") as counter:
         for record in my_records:
             # Do something with the record
             counter.increment()

    Args:
        stream: The stream name.
        endpoint: The endpoint name.
        log_interval: The interval at which to log the count.
        tags: Tags to add to the measurement.

    Returns:
        A counter for counting records.
    """
    tags[Tag.STREAM] = stream
    if endpoint:
        tags[Tag.ENDPOINT] = endpoint
    return Counter(Metric.RECORD_COUNT, tags, log_interval=log_interval)


def batch_counter(stream: str, **tags: t.Any) -> Counter:
    """Use for counting batches sent to the target.

    with batch_counter("my_stream") as counter:
         for batch in my_batches:
             # Do something with the batch
             counter.increment()

    Args:
        stream: The stream name.
        tags: Tags to add to the measurement.

    Returns:
        A counter for counting batches.
    """
    tags[Tag.STREAM] = stream
    return Counter(Metric.BATCH_COUNT, tags)


def http_request_counter(
    stream: str,
    endpoint: str,
    log_interval: float = DEFAULT_LOG_INTERVAL,
    **tags: t.Any,
) -> Counter:
    """Use for counting HTTP requests.

    with http_request_counter() as counter:
         for record in my_records:
             # Do something with the record
             counter.increment()

    Args:
        stream: The stream name.
        endpoint: The endpoint name.
        log_interval: The interval at which to log the count.
        tags: Tags to add to the measurement.

    Returns:
        A counter for counting HTTP requests.
    """
    tags.update({Tag.STREAM: stream, Tag.ENDPOINT: endpoint})
    return Counter(Metric.HTTP_REQUEST_COUNT, tags, log_interval=log_interval)


def sync_timer(stream: str, **tags: t.Any) -> Timer:
    """Use for timing the sync of a stream.

    with singer.metrics.sync_timer() as timer:
         # Do something
         print(f"Sync took {timer.elapsed()} seconds")

    Args:
        stream: The stream name.
        tags: Tags to add to the measurement.

    Returns:
        A timer for timing the sync of a stream.
    """
    tags[Tag.STREAM] = stream
    return Timer(Metric.SYNC_DURATION, tags)


def _load_yaml_logging_config(path: Traversable | Path) -> t.Any:  # noqa: ANN401
    """Load the logging config from the YAML file.

    Args:
        path: A path to the YAML file.

    Returns:
        The logging config.
    """
    with path.open() as f:
        return yaml.safe_load(f)


def _get_default_config_path(package: str) -> Traversable:
    """Get a logging configuration.

    Args:
        package: The package name to get the logging configuration for.

    Returns:
        A logging configuration.
    """
    filename = "default_logging.yml"
    path = get_package_files(package) / filename
    return path if path.is_file() else get_package_files("singer_sdk") / filename


def _setup_logging(config: t.Mapping[str, t.Any], *, package: str) -> None:
    """Setup logging.

    Args:
        package: The package name to get the logging configuration for.
        config: A plugin configuration dictionary.
    """
    path = _get_default_config_path(package)
    logging.config.dictConfig(_load_yaml_logging_config(path))

    config = config or {}
    metrics_log_level = config.get(METRICS_LOG_LEVEL_SETTING, "INFO").upper()
    logging.getLogger(METRICS_LOGGER_NAME).setLevel(metrics_log_level)

    if "SINGER_SDK_LOG_CONFIG" in os.environ:
        log_config_path = Path(os.environ["SINGER_SDK_LOG_CONFIG"])
        logging.config.dictConfig(_load_yaml_logging_config(log_config_path))



"""Tools and standard tests for Tap/Target implementations."""

from __future__ import annotations

import typing as t
import warnings

from .config import SuiteConfig
from .factory import get_tap_test_class, get_target_test_class
from .legacy import (
    _get_tap_catalog,
    _select_all,
    sync_end_to_end,
    tap_sync_test,
    tap_to_target_sync_test,
    target_sync_test,
)
from .runners import SingerTestRunner, TapTestRunner, TargetTestRunner


def __getattr__(name: str) -> t.Any:  # noqa: ANN401
    if name == "get_standard_tap_tests":
        warnings.warn(
            "The function singer_sdk.testing.get_standard_tap_tests is deprecated "
            "and will be removed in a future release. Use get_tap_test_class instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        from .legacy import get_standard_tap_tests  # noqa: PLC0415

        return get_standard_tap_tests

    if name == "get_standard_target_tests":
        warnings.warn(
            "The function singer_sdk.testing.get_standard_target_tests is deprecated "
            "and will be removed in a future release. Use get_target_test_class "
            "instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        from .legacy import get_standard_target_tests  # noqa: PLC0415

        return get_standard_target_tests

    msg = f"module {__name__} has no attribute {name}"
    raise AttributeError(msg)


__all__ = [
    "SingerTestRunner",
    "SuiteConfig",
    "TapTestRunner",
    "TargetTestRunner",
    "_get_tap_catalog",
    "_select_all",
    "get_tap_test_class",
    "get_target_test_class",
    "sync_end_to_end",
    "tap_sync_test",
    "tap_to_target_sync_test",
    "target_sync_test",
]
